@@ -26,6 +26,7 @@ from osc_lib import utils
 from openstackclient.i18n import _
 from openstackclient.identity import common as identity_common
 from openstackclient.network import sdk_utils
+from openstackclient.network.v2 import _tag
 
 
 LOG = logging.getLogger(__name__)
@@ -35,10 +36,6 @@ def _format_admin_state(state):
     return 'UP' if state else 'DOWN'
 
 
-def _format_dns_assignment(dns_assignment):
-    return utils.format_list_of_dicts(dns_assignment) \
-        if dns_assignment else None
-
 _formatters = {
     'admin_state_up': _format_admin_state,
     'is_admin_state_up': _format_admin_state,
@@ -47,10 +44,11 @@ _formatters = {
     'binding_vif_details': utils.format_dict,
     'binding:profile': utils.format_dict,
     'binding:vif_details': utils.format_dict,
-    'dns_assignment': _format_dns_assignment,
+    'dns_assignment': utils.format_list_of_dicts,
     'extra_dhcp_opts': utils.format_list_of_dicts,
     'fixed_ips': utils.format_list_of_dicts,
     'security_group_ids': utils.format_list,
+    'tags': utils.format_list,
 }
 
 
@@ -114,8 +112,6 @@ def _get_attrs(client_manager, parsed_args):
         ))
     if parsed_args.description is not None:
         attrs['description'] = parsed_args.description
-    if parsed_args.fixed_ip is not None:
-        attrs['fixed_ips'] = parsed_args.fixed_ip
     if parsed_args.device:
         attrs['device_id'] = parsed_args.device
     if parsed_args.device_owner is not None:
@@ -124,12 +120,12 @@ def _get_attrs(client_manager, parsed_args):
         attrs['admin_state_up'] = True
     if parsed_args.disable:
         attrs['admin_state_up'] = False
-    if parsed_args.binding_profile is not None:
-        attrs['binding:profile'] = parsed_args.binding_profile
     if parsed_args.vnic_type is not None:
         attrs['binding:vnic_type'] = parsed_args.vnic_type
     if parsed_args.host:
         attrs['binding:host_id'] = parsed_args.host
+    if parsed_args.mac_address is not None:
+        attrs['mac_address'] = parsed_args.mac_address
 
     if parsed_args.dns_name is not None:
         attrs['dns_name'] = parsed_args.dns_name
@@ -138,8 +134,6 @@ def _get_attrs(client_manager, parsed_args):
         attrs['name'] = str(parsed_args.name)
     # The remaining options do not support 'port set' command, so they require
     # additional check
-    if 'mac_address' in parsed_args and parsed_args.mac_address is not None:
-        attrs['mac_address'] = parsed_args.mac_address
     if 'network' in parsed_args and parsed_args.network is not None:
         attrs['network_id'] = parsed_args.network
     if 'project' in parsed_args and parsed_args.project is not None:
@@ -159,6 +153,13 @@ def _get_attrs(client_manager, parsed_args):
     if parsed_args.enable_port_security:
         attrs['port_security_enabled'] = True
 
+    if 'no_qos_policy' in parsed_args and parsed_args.no_qos_policy:
+        attrs['qos_policy_id'] = None
+
+    if parsed_args.qos_policy:
+        attrs['qos_policy_id'] = client_manager.network.find_qos_policy(
+            parsed_args.qos_policy, ignore_missing=False).id
+
     return attrs
 
 
@@ -234,6 +235,11 @@ def _add_updatable_args(parser):
         metavar='<device-id>',
         help=argparse.SUPPRESS,
     )
+    parser.add_argument(
+        '--mac-address',
+        metavar='<mac-address>',
+        help=_("MAC address of this port (admin only)")
+    )
     parser.add_argument(
         '--device-owner',
         metavar='<device-owner>',
@@ -244,9 +250,10 @@ def _add_updatable_args(parser):
         '--vnic-type',
         metavar='<vnic-type>',
         choices=['direct', 'direct-physical', 'macvtap',
-                 'normal', 'baremetal'],
+                 'normal', 'baremetal', 'virtio-forwarder'],
         help=_("VNIC type for this port (direct | direct-physical | "
-               "macvtap | normal | baremetal, default: normal)")
+               "macvtap | normal | baremetal | virtio-forwarder, "
+               " default: normal)")
     )
     # NOTE(dtroyer): --host-id is deprecated in Mar 2016.  Do not
     #                remove before 3.x release or Mar 2017.
@@ -295,15 +302,21 @@ class CreatePort(command.ShowOne):
             help=_("Network this port belongs to (name or ID)")
         )
         _add_updatable_args(parser)
-        parser.add_argument(
+        fixed_ip = parser.add_mutually_exclusive_group()
+        fixed_ip.add_argument(
             '--fixed-ip',
             metavar='subnet=<subnet>,ip-address=<ip-address>',
             action=parseractions.MultiKeyValueAction,
             optional_keys=['subnet', 'ip-address'],
-            help=_("Desired IP and/or subnet (name or ID) for this port: "
+            help=_("Desired IP and/or subnet for this port (name or ID): "
                    "subnet=<subnet>,ip-address=<ip-address> "
                    "(repeat option to set multiple fixed IP addresses)")
         )
+        fixed_ip.add_argument(
+            '--no-fixed-ip',
+            action='store_true',
+            help=_("No IP or subnet for this port.")
+        )
         parser.add_argument(
             '--binding-profile',
             metavar='<binding-profile>',
@@ -324,11 +337,6 @@ class CreatePort(command.ShowOne):
             action='store_true',
             help=_("Disable port")
         )
-        parser.add_argument(
-            '--mac-address',
-            metavar='<mac-address>',
-            help=_("MAC address of this port")
-        )
         parser.add_argument(
             '--project',
             metavar='<project>',
@@ -341,7 +349,7 @@ class CreatePort(command.ShowOne):
             help=_("Name of this port")
         )
         # TODO(singhj): Add support for extended options:
-        # qos,dhcp
+        # dhcp
         secgroups = parser.add_mutually_exclusive_group()
         secgroups.add_argument(
             '--security-group',
@@ -357,6 +365,11 @@ class CreatePort(command.ShowOne):
             action='store_true',
             help=_("Associate no security groups with this port")
         )
+        parser.add_argument(
+            '--qos-policy',
+            metavar='<qos-policy>',
+            help=_("Attach QoS policy to this port (name or ID)")
+        )
         port_security = parser.add_mutually_exclusive_group()
         port_security.add_argument(
             '--enable-port-security',
@@ -379,6 +392,7 @@ class CreatePort(command.ShowOne):
                    "ip-address=<ip-address>[,mac-address=<mac-address>] "
                    "(repeat option to set multiple allowed-address pairs)")
         )
+        _tag.add_tag_option_to_parser_for_create(parser, _('port'))
         return parser
 
     def take_action(self, parsed_args):
@@ -389,6 +403,14 @@ class CreatePort(command.ShowOne):
         _prepare_fixed_ips(self.app.client_manager, parsed_args)
         attrs = _get_attrs(self.app.client_manager, parsed_args)
 
+        if parsed_args.binding_profile is not None:
+            attrs['binding:profile'] = parsed_args.binding_profile
+
+        if parsed_args.fixed_ip:
+            attrs['fixed_ips'] = parsed_args.fixed_ip
+        elif parsed_args.no_fixed_ip:
+            attrs['fixed_ips'] = []
+
         if parsed_args.security_group:
             attrs['security_group_ids'] = [client.find_security_group(
                                            sg, ignore_missing=False).id
@@ -396,11 +418,17 @@ class CreatePort(command.ShowOne):
                                            parsed_args.security_group]
         elif parsed_args.no_security_group:
             attrs['security_group_ids'] = []
+
         if parsed_args.allowed_address_pairs:
             attrs['allowed_address_pairs'] = (
                 _convert_address_pairs(parsed_args))
 
+        if parsed_args.qos_policy:
+            attrs['qos_policy_id'] = client.find_qos_policy(
+                parsed_args.qos_policy, ignore_missing=False).id
         obj = client.create_port(**attrs)
+        # tags cannot be set when created, so tags need to be set later.
+        _tag.update_tags_for_set(client, obj, parsed_args)
         display_columns, columns = _get_columns(obj)
         data = utils.get_item_properties(obj, columns, formatters=_formatters)
 
@@ -493,15 +521,15 @@ class ListPort(command.Lister):
             metavar='subnet=<subnet>,ip-address=<ip-address>',
             action=parseractions.MultiKeyValueAction,
             optional_keys=['subnet', 'ip-address'],
-            help=_("Desired IP and/or subnet (name or ID) for filtering "
-                   "ports: subnet=<subnet>,ip-address=<ip-address> "
-                   "(repeat option to set multiple fixed IP addresses)")
+            help=_("Desired IP and/or subnet for filtering ports "
+                   "(name or ID): subnet=<subnet>,ip-address=<ip-address> "
+                   "(repeat option to set multiple fixed IP addresses)"),
         )
+        _tag.add_tag_filtering_option_to_parser(parser, _('ports'))
         return parser
 
     def take_action(self, parsed_args):
         network_client = self.app.client_manager.network
-        compute_client = self.app.client_manager.compute
         identity_client = self.app.client_manager.identity
 
         columns = (
@@ -521,8 +549,8 @@ class ListPort(command.Lister):
 
         filters = {}
         if parsed_args.long:
-            columns += ('security_group_ids', 'device_owner',)
-            column_headers += ('Security Groups', 'Device Owner',)
+            columns += ('security_group_ids', 'device_owner', 'tags')
+            column_headers += ('Security Groups', 'Device Owner', 'Tags')
         if parsed_args.device_owner is not None:
             filters['device_owner'] = parsed_args.device_owner
         if parsed_args.router:
@@ -530,6 +558,7 @@ class ListPort(command.Lister):
                                                  ignore_missing=False)
             filters['device_id'] = _router.id
         if parsed_args.server:
+            compute_client = self.app.client_manager.compute
             server = utils.find_resource(compute_client.servers,
                                          parsed_args.server)
             filters['device_id'] = server.id
@@ -551,6 +580,8 @@ class ListPort(command.Lister):
             filters['fixed_ips'] = _prepare_filter_fixed_ips(
                 self.app.client_manager, parsed_args)
 
+        _tag.get_tag_filtering_args(parsed_args, filters)
+
         data = network_client.ports(**filters)
 
         return (column_headers,
@@ -590,7 +621,7 @@ class SetPort(command.Command):
             metavar='subnet=<subnet>,ip-address=<ip-address>',
             action=parseractions.MultiKeyValueAction,
             optional_keys=['subnet', 'ip-address'],
-            help=_("Desired IP and/or subnet (name or ID) for this port: "
+            help=_("Desired IP and/or subnet for this port (name or ID): "
                    "subnet=<subnet>,ip-address=<ip-address> "
                    "(repeat option to set multiple fixed IP addresses)")
         )
@@ -616,6 +647,11 @@ class SetPort(command.Command):
                    "Specify both --binding-profile and --no-binding-profile "
                    "to overwrite the current binding:profile information.")
         )
+        parser.add_argument(
+            '--qos-policy',
+            metavar='<qos-policy>',
+            help=_("Attach QoS policy to this port (name or ID)")
+        )
         parser.add_argument(
             'port',
             metavar="<port>",
@@ -666,58 +702,76 @@ class SetPort(command.Command):
                    "(Specify both --allowed-address and --no-allowed-address"
                    "to overwrite the current allowed-address pairs)")
         )
+        parser.add_argument(
+            '--data-plane-status',
+            metavar='<status>',
+            choices=['ACTIVE', 'DOWN'],
+            help=_("Set data plane status of this port (ACTIVE | DOWN). "
+                   "Unset it to None with the 'port unset' command "
+                   "(requires data plane status extension)")
+        )
+        _tag.add_tag_option_to_parser_for_set(parser, _('port'))
+
         return parser
 
     def take_action(self, parsed_args):
         client = self.app.client_manager.network
 
         _prepare_fixed_ips(self.app.client_manager, parsed_args)
-        attrs = _get_attrs(self.app.client_manager, parsed_args)
         obj = client.find_port(parsed_args.port, ignore_missing=False)
-        if 'binding:profile' in attrs:
-            # Do not modify attrs if both binding_profile/no_binding given
-            if not parsed_args.no_binding_profile:
-                tmp_binding_profile = copy.deepcopy(obj.binding_profile)
-                tmp_binding_profile.update(attrs['binding:profile'])
-                attrs['binding:profile'] = tmp_binding_profile
-        elif parsed_args.no_binding_profile:
+        attrs = _get_attrs(self.app.client_manager, parsed_args)
+
+        if parsed_args.no_binding_profile:
             attrs['binding:profile'] = {}
-        if 'fixed_ips' in attrs:
-            # When user unsets the fixed_ips, obj.fixed_ips = [{}].
-            # Adding the obj.fixed_ips list to attrs['fixed_ips']
-            # would therefore add an empty dictionary, while we need
-            # to append the attrs['fixed_ips'] iff there is some info
-            # in the obj.fixed_ips. Therefore I have opted for this `for` loop
-            # Do not modify attrs if fixed_ip/no_fixed_ip given
-            if not parsed_args.no_fixed_ip:
-                attrs['fixed_ips'] += [ip for ip in obj.fixed_ips if ip]
-        elif parsed_args.no_fixed_ip:
+        if parsed_args.binding_profile:
+            if 'binding:profile' not in attrs:
+                attrs['binding:profile'] = copy.deepcopy(obj.binding_profile)
+            attrs['binding:profile'].update(parsed_args.binding_profile)
+
+        if parsed_args.no_fixed_ip:
             attrs['fixed_ips'] = []
+        if parsed_args.fixed_ip:
+            if 'fixed_ips' not in attrs:
+                # obj.fixed_ips = [{}] if no fixed IPs are set.
+                # Only append this to attrs['fixed_ips'] if actual fixed
+                # IPs are present to avoid adding an empty dict.
+                attrs['fixed_ips'] = [ip for ip in obj.fixed_ips if ip]
+            attrs['fixed_ips'].extend(parsed_args.fixed_ip)
 
-        if parsed_args.security_group:
-            attrs['security_group_ids'] = [
-                client.find_security_group(sg, ignore_missing=False).id for
-                sg in parsed_args.security_group]
-            if not parsed_args.no_security_group:
-                attrs['security_group_ids'] += obj.security_group_ids
-
-        elif parsed_args.no_security_group:
+        if parsed_args.no_security_group:
             attrs['security_group_ids'] = []
+        if parsed_args.security_group:
+            if 'security_group_ids' not in attrs:
+                # NOTE(dtroyer): Get existing security groups, iterate the
+                #                list to force a new list object to be
+                #                created and make sure the SDK Resource
+                #                marks the attribute 'dirty'.
+                attrs['security_group_ids'] = [
+                    id for id in obj.security_group_ids
+                ]
+            attrs['security_group_ids'].extend(
+                client.find_security_group(sg, ignore_missing=False).id
+                for sg in parsed_args.security_group
+            )
 
-        if (parsed_args.allowed_address_pairs and
-                parsed_args.no_allowed_address_pair):
-            attrs['allowed_address_pairs'] = (
-                _convert_address_pairs(parsed_args))
-
-        elif parsed_args.allowed_address_pairs:
-            attrs['allowed_address_pairs'] = (
-                [addr for addr in obj.allowed_address_pairs if addr] +
-                _convert_address_pairs(parsed_args))
-
-        elif parsed_args.no_allowed_address_pair:
+        if parsed_args.no_allowed_address_pair:
             attrs['allowed_address_pairs'] = []
+        if parsed_args.allowed_address_pairs:
+            if 'allowed_address_pairs' not in attrs:
+                attrs['allowed_address_pairs'] = (
+                    [addr for addr in obj.allowed_address_pairs if addr]
+                )
+            attrs['allowed_address_pairs'].extend(
+                _convert_address_pairs(parsed_args)
+            )
+        if parsed_args.data_plane_status:
+            attrs['data_plane_status'] = parsed_args.data_plane_status
 
-        client.update_port(obj, **attrs)
+        if attrs:
+            client.update_port(obj, **attrs)
+
+        # tags is a subresource and it needs to be updated separately.
+        _tag.update_tags_for_set(client, obj, parsed_args)
 
 
 class ShowPort(command.ShowOne):
@@ -752,8 +806,8 @@ class UnsetPort(command.Command):
             metavar='subnet=<subnet>,ip-address=<ip-address>',
             action=parseractions.MultiKeyValueAction,
             optional_keys=['subnet', 'ip-address'],
-            help=_("Desired IP and/or subnet (name or ID) which should be "
-                   "removed from this port: subnet=<subnet>,"
+            help=_("Desired IP and/or subnet which should be "
+                   "removed from this port (name or ID): subnet=<subnet>,"
                    "ip-address=<ip-address> (repeat option to unset multiple "
                    "fixed IP addresses)"))
 
@@ -789,6 +843,19 @@ class UnsetPort(command.Command):
                    "[,mac-address=<mac-address>] (repeat option to set "
                    "multiple allowed-address pairs)")
         )
+        parser.add_argument(
+            '--qos-policy',
+            action='store_true',
+            default=False,
+            help=_("Remove the QoS policy attached to the port")
+        )
+        parser.add_argument(
+            '--data-plane-status',
+            action='store_true',
+            help=_("Clear existing information of data plane status")
+        )
+
+        _tag.add_tag_option_to_parser_for_unset(parser, _('port'))
 
         return parser
 
@@ -838,6 +905,13 @@ class UnsetPort(command.Command):
                 msg = _("Port does not contain allowed-address-pair %s") % addr
                 raise exceptions.CommandError(msg)
             attrs['allowed_address_pairs'] = tmp_addr_pairs
+        if parsed_args.qos_policy:
+            attrs['qos_policy_id'] = None
+        if parsed_args.data_plane_status:
+            attrs['data_plane_status'] = None
 
         if attrs:
             client.update_port(obj, **attrs)
+
+        # tags is a subresource and it needs to be updated separately.
+        _tag.update_tags_for_unset(client, obj, parsed_args)
