@@ -16,11 +16,6 @@
 import argparse
 import logging
 
-try:
-    from novaclient.v2 import security_group_rules as compute_secgroup_rules
-except ImportError:
-    from novaclient.v1_1 import security_group_rules as compute_secgroup_rules
-
 from osc_lib.cli import parseractions
 from osc_lib import exceptions
 from osc_lib import utils
@@ -164,8 +159,8 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
             help=_("IP protocol (ah, dccp, egp, esp, gre, icmp, igmp, "
                    "ipv6-encap, ipv6-frag, ipv6-icmp, ipv6-nonxt, "
                    "ipv6-opts, ipv6-route, ospf, pgm, rsvp, sctp, tcp, "
-                   "udp, udplite, vrrp and integer representations [0-255]; "
-                   "default: tcp)")
+                   "udp, udplite, vrrp and integer representations [0-255] "
+                   "or any; default: tcp)")
         )
         protocol_group.add_argument(
             '--proto',
@@ -235,6 +230,8 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
             protocol = parsed_args.protocol
         if parsed_args.proto is not None:
             protocol = parsed_args.proto
+        if protocol == 'any':
+            protocol = None
         return protocol
 
     def _is_ipv6_protocol(self, protocol):
@@ -242,7 +239,7 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
         # However, while the OSC CLI doesn't document the protocol,
         # the code must still handle it. In addition, handle both
         # protocol names and numbers.
-        if (protocol.startswith('ipv6-') or
+        if (protocol is not None and protocol.startswith('ipv6-') or
                 protocol in ['icmpv6', '41', '43', '44', '58', '59', '60']):
             return True
         else:
@@ -299,9 +296,9 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
         if parsed_args.dst_port and not is_icmp_protocol:
             attrs['port_range_min'] = parsed_args.dst_port[0]
             attrs['port_range_max'] = parsed_args.dst_port[1]
-        if parsed_args.icmp_type:
+        if parsed_args.icmp_type is not None and parsed_args.icmp_type >= 0:
             attrs['port_range_min'] = parsed_args.icmp_type
-        if parsed_args.icmp_code:
+        if parsed_args.icmp_code is not None and parsed_args.icmp_code >= 0:
             attrs['port_range_max'] = parsed_args.icmp_code
 
         # NOTE(dtroyer): --src-ip and --src-group were deprecated in Nov 2016.
@@ -348,10 +345,7 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
         return (display_columns, data)
 
     def take_action_compute(self, client, parsed_args):
-        group = utils.find_resource(
-            client.security_groups,
-            parsed_args.group,
-        )
+        group = client.api.security_group_find(parsed_args.group)
         protocol = self._get_protocol(parsed_args)
         if protocol == 'icmp':
             from_port, to_port = -1, -1
@@ -363,10 +357,9 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
         remote_ip = None
         if not (parsed_args.remote_group is None and
                 parsed_args.src_group is None):
-            parsed_args.remote_group = utils.find_resource(
-                client.security_groups,
+            parsed_args.remote_group = client.api.security_group_find(
                 parsed_args.remote_group or parsed_args.src_group,
-            ).id
+            )['id']
             if parsed_args.src_group:
                 LOG.warning(
                     _("The %(old)s option is deprecated, "
@@ -384,15 +377,16 @@ class CreateSecurityGroupRule(common.NetworkAndComputeShowOne):
                 )
         else:
             remote_ip = '0.0.0.0/0'
-        obj = client.security_group_rules.create(
-            group.id,
-            protocol,
-            from_port,
-            to_port,
-            remote_ip,
-            parsed_args.remote_group,
+
+        obj = client.api.security_group_rule_create(
+            security_group_id=group['id'],
+            ip_protocol=protocol,
+            from_port=from_port,
+            to_port=to_port,
+            remote_ip=remote_ip,
+            remote_group=parsed_args.remote_group,
         )
-        return _format_security_group_rule_show(obj._info)
+        return _format_security_group_rule_show(obj)
 
 
 class DeleteSecurityGroupRule(common.NetworkAndComputeDelete):
@@ -417,7 +411,7 @@ class DeleteSecurityGroupRule(common.NetworkAndComputeDelete):
         client.delete_security_group_rule(obj)
 
     def take_action_compute(self, client, parsed_args):
-        client.security_group_rules.delete(self.r)
+        client.api.security_group_rule_delete(self.r)
 
 
 class ListSecurityGroupRule(common.NetworkAndComputeLister):
@@ -567,27 +561,29 @@ class ListSecurityGroupRule(common.NetworkAndComputeLister):
 
         rules_to_list = []
         if parsed_args.group is not None:
-            group = utils.find_resource(
-                client.security_groups,
+            group = client.api.security_group_find(
                 parsed_args.group,
             )
-            rules_to_list = group.rules
+            rules_to_list = group['rules']
         else:
             columns = columns + ('parent_group_id',)
             search = {'all_tenants': parsed_args.all_projects}
-            for group in client.security_groups.list(search_opts=search):
-                rules_to_list.extend(group.rules)
+            for group in client.api.security_group_list(search_opts=search):
+                rules_to_list.extend(group['rules'])
 
         # NOTE(rtheis): Turn the raw rules into resources.
         rules = []
         for rule in rules_to_list:
-            rules.append(compute_secgroup_rules.SecurityGroupRule(
-                client.security_group_rules,
+            rules.append(
                 network_utils.transform_compute_security_group_rule(rule),
-            ))
+            )
+            # rules.append(compute_secgroup_rules.SecurityGroupRule(
+            #     client.security_group_rules,
+            #     network_utils.transform_compute_security_group_rule(rule),
+            # ))
 
         return (column_headers,
-                (utils.get_item_properties(
+                (utils.get_dict_properties(
                     s, columns,
                 ) for s in rules))
 
@@ -617,8 +613,8 @@ class ShowSecurityGroupRule(common.NetworkAndComputeShowOne):
         # the requested rule.
         obj = None
         security_group_rules = []
-        for security_group in client.security_groups.list():
-            security_group_rules.extend(security_group.rules)
+        for security_group in client.api.security_group_list():
+            security_group_rules.extend(security_group['rules'])
         for security_group_rule in security_group_rules:
             if parsed_args.rule == str(security_group_rule.get('id')):
                 obj = security_group_rule
