@@ -129,6 +129,7 @@ class TestCreateRouter(TestRouter):
         'project_id',
         'routes',
         'status',
+        'tags',
     )
     data = (
         router._format_admin_state(new_router.admin_state_up),
@@ -136,19 +137,21 @@ class TestCreateRouter(TestRouter):
         osc_utils.format_list(new_router.availability_zones),
         new_router.description,
         new_router.distributed,
-        router._format_external_gateway_info(new_router.external_gateway_info),
+        router._format_router_info(new_router.external_gateway_info),
         new_router.ha,
         new_router.id,
         new_router.name,
         new_router.tenant_id,
         router._format_routes(new_router.routes),
         new_router.status,
+        osc_utils.format_list(new_router.tags),
     )
 
     def setUp(self):
         super(TestCreateRouter, self).setUp()
 
         self.network.create_router = mock.Mock(return_value=self.new_router)
+        self.network.set_tags = mock.Mock(return_value=None)
 
         # Get the command object to test
         self.cmd = router.CreateRouter(self.app, self.namespace)
@@ -159,6 +162,7 @@ class TestCreateRouter(TestRouter):
 
         self.assertRaises(tests_utils.ParserException, self.check_parser,
                           self.cmd, arglist, verifylist)
+        self.assertFalse(self.network.set_tags.called)
 
     def test_create_default_options(self):
         arglist = [
@@ -178,19 +182,21 @@ class TestCreateRouter(TestRouter):
             'admin_state_up': True,
             'name': self.new_router.name,
         })
+        self.assertFalse(self.network.set_tags.called)
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, data)
 
-    def test_create_with_ha_option(self):
+    def _test_create_with_ha_options(self, option, ha):
         arglist = [
-            '--ha',
+            option,
             self.new_router.name,
         ]
         verifylist = [
             ('name', self.new_router.name),
             ('enable', True),
             ('distributed', False),
-            ('ha', True),
+            ('ha', ha),
+            ('no_ha', not ha),
         ]
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
@@ -199,11 +205,46 @@ class TestCreateRouter(TestRouter):
         self.network.create_router.assert_called_once_with(**{
             'admin_state_up': True,
             'name': self.new_router.name,
-            'ha': True,
+            'ha': ha,
         })
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, data)
 
+    def test_create_with_ha_option(self):
+        self._test_create_with_ha_options('--ha', True)
+
+    def test_create_with_no_ha_option(self):
+        self._test_create_with_ha_options('--no-ha', False)
+
+    def _test_create_with_distributed_options(self, option, distributed):
+        arglist = [
+            option,
+            self.new_router.name,
+        ]
+        verifylist = [
+            ('name', self.new_router.name),
+            ('enable', True),
+            ('distributed', distributed),
+            ('centralized', not distributed),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        columns, data = (self.cmd.take_action(parsed_args))
+
+        self.network.create_router.assert_called_once_with(**{
+            'admin_state_up': True,
+            'name': self.new_router.name,
+            'distributed': distributed,
+        })
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, data)
+
+    def test_create_with_distributed_option(self):
+        self._test_create_with_distributed_options('--distributed', True)
+
+    def test_create_with_centralized_option(self):
+        self._test_create_with_distributed_options('--centralized', False)
+
     def test_create_with_AZ_hints(self):
         arglist = [
             self.new_router.name,
@@ -229,6 +270,45 @@ class TestCreateRouter(TestRouter):
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, data)
 
+    def _test_create_with_tag(self, add_tags=True):
+        arglist = [self.new_router.name]
+        if add_tags:
+            arglist += ['--tag', 'red', '--tag', 'blue']
+        else:
+            arglist += ['--no-tag']
+        verifylist = [
+            ('name', self.new_router.name),
+            ('enable', True),
+            ('distributed', False),
+            ('ha', False),
+        ]
+        if add_tags:
+            verifylist.append(('tags', ['red', 'blue']))
+        else:
+            verifylist.append(('no_tag', True))
+
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        columns, data = (self.cmd.take_action(parsed_args))
+
+        self.network.create_router.assert_called_once_with(
+            name=self.new_router.name,
+            admin_state_up=True
+        )
+        if add_tags:
+            self.network.set_tags.assert_called_once_with(
+                self.new_router,
+                tests_utils.CompareBySet(['red', 'blue']))
+        else:
+            self.assertFalse(self.network.set_tags.called)
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, data)
+
+    def test_create_with_tags(self):
+        self._test_create_with_tag(add_tags=True)
+
+    def test_create_with_no_tag(self):
+        self._test_create_with_tag(add_tags=False)
+
 
 class TestDeleteRouter(TestRouter):
 
@@ -327,11 +407,13 @@ class TestListRouter(TestRouter):
     columns_long = columns + (
         'Routes',
         'External gateway info',
-        'Availability zones'
+        'Availability zones',
+        'Tags',
     )
     columns_long_no_az = columns + (
         'Routes',
         'External gateway info',
+        'Tags',
     )
 
     data = []
@@ -345,14 +427,30 @@ class TestListRouter(TestRouter):
             r.ha,
             r.tenant_id,
         ))
+
+    router_agent_data = []
+    for r in routers:
+        router_agent_data.append((
+            r.id,
+            r.name,
+            r.external_gateway_info,
+        ))
+
+    agents_columns = (
+        'ID',
+        'Name',
+        'External Gateway Info',
+    )
+
     data_long = []
     for i in range(0, len(routers)):
         r = routers[i]
         data_long.append(
             data[i] + (
                 router._format_routes(r.routes),
-                router._format_external_gateway_info(r.external_gateway_info),
+                router._format_router_info(r.external_gateway_info),
                 osc_utils.format_list(r.availability_zones),
+                osc_utils.format_list(r.tags),
             )
         )
     data_long_no_az = []
@@ -361,7 +459,8 @@ class TestListRouter(TestRouter):
         data_long_no_az.append(
             data[i] + (
                 router._format_routes(r.routes),
-                router._format_external_gateway_info(r.external_gateway_info),
+                router._format_router_info(r.external_gateway_info),
+                osc_utils.format_list(r.tags),
             )
         )
 
@@ -371,8 +470,15 @@ class TestListRouter(TestRouter):
         # Get the command object to test
         self.cmd = router.ListRouter(self.app, self.namespace)
 
+        self.network.agent_hosted_routers = mock.Mock(
+            return_value=self.routers)
         self.network.routers = mock.Mock(return_value=self.routers)
         self.network.find_extension = mock.Mock(return_value=self._extensions)
+        self.network.find_router = mock.Mock(return_value=self.routers[0])
+        self._testagent = \
+            network_fakes.FakeNetworkAgent.create_one_network_agent()
+        self.network.get_agent = mock.Mock(return_value=self._testagent)
+        self.network.get_router = mock.Mock(return_value=self.routers[0])
 
     def test_router_list_no_options(self):
         arglist = []
@@ -459,7 +565,7 @@ class TestListRouter(TestRouter):
         columns, data = self.cmd.take_action(parsed_args)
 
         self.network.routers.assert_called_once_with(
-            **{'admin_state_up': True}
+            **{'admin_state_up': True, 'is_admin_state_up': True}
         )
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, list(data))
@@ -476,7 +582,7 @@ class TestListRouter(TestRouter):
         columns, data = self.cmd.take_action(parsed_args)
 
         self.network.routers.assert_called_once_with(
-            **{'admin_state_up': False}
+            **{'admin_state_up': False, 'is_admin_state_up': False}
         )
 
         self.assertEqual(self.columns, columns)
@@ -494,7 +600,7 @@ class TestListRouter(TestRouter):
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
         columns, data = self.cmd.take_action(parsed_args)
-        filters = {'tenant_id': project.id}
+        filters = {'tenant_id': project.id, 'project_id': project.id}
 
         self.network.routers.assert_called_once_with(**filters)
         self.assertEqual(self.columns, columns)
@@ -514,12 +620,65 @@ class TestListRouter(TestRouter):
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
         columns, data = self.cmd.take_action(parsed_args)
-        filters = {'tenant_id': project.id}
+        filters = {'tenant_id': project.id, 'project_id': project.id}
 
         self.network.routers.assert_called_once_with(**filters)
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, list(data))
 
+    def test_router_list_agents_no_args(self):
+        arglist = [
+            '--agents',
+        ]
+        verifylist = []
+
+        # Missing required router ID should bail here
+        self.assertRaises(tests_utils.ParserException, self.check_parser,
+                          self.cmd, arglist, verifylist)
+
+    def test_router_list_agents(self):
+        arglist = [
+            '--agent', self._testagent.id,
+        ]
+        verifylist = [
+            ('agent', self._testagent.id),
+        ]
+
+        attrs = {self._testagent.id, }
+
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.network.agent_hosted_routers(
+            *attrs)
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, list(data))
+
+    def test_list_with_tag_options(self):
+        arglist = [
+            '--tags', 'red,blue',
+            '--any-tags', 'red,green',
+            '--not-tags', 'orange,yellow',
+            '--not-any-tags', 'black,white',
+        ]
+        verifylist = [
+            ('tags', ['red', 'blue']),
+            ('any_tags', ['red', 'green']),
+            ('not_tags', ['orange', 'yellow']),
+            ('not_any_tags', ['black', 'white']),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        columns, data = self.cmd.take_action(parsed_args)
+
+        self.network.routers.assert_called_once_with(
+            **{'tags': 'red,blue',
+               'any_tags': 'red,green',
+               'not_tags': 'orange,yellow',
+               'not_any_tags': 'black,white'}
+        )
+        self.assertEqual(self.columns, columns)
+        self.assertEqual(self.data, list(data))
+
 
 class TestRemovePortFromRouter(TestRouter):
     '''Remove port from a Router '''
@@ -602,17 +761,21 @@ class TestSetRouter(TestRouter):
 
     # The router to set.
     _default_route = {'destination': '10.20.20.0/24', 'nexthop': '10.20.30.1'}
+    _network = network_fakes.FakeNetwork.create_one_network()
+    _subnet = network_fakes.FakeSubnet.create_one_subnet()
     _router = network_fakes.FakeRouter.create_one_router(
-        attrs={'routes': [_default_route]}
+        attrs={'routes': [_default_route],
+               'tags': ['green', 'red']}
     )
 
     def setUp(self):
         super(TestSetRouter, self).setUp()
-
+        self.network.router_add_gateway = mock.Mock()
         self.network.update_router = mock.Mock(return_value=None)
-
+        self.network.set_tags = mock.Mock(return_value=None)
         self.network.find_router = mock.Mock(return_value=self._router)
-
+        self.network.find_network = mock.Mock(return_value=self._network)
+        self.network.find_subnet = mock.Mock(return_value=self._subnet)
         # Get the command object to test
         self.cmd = router.SetRouter(self.app, self.namespace)
 
@@ -702,10 +865,10 @@ class TestSetRouter(TestRouter):
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
 
         result = self.cmd.take_action(parsed_args)
-
+        routes = [{'destination': '10.20.30.0/24',
+                   'nexthop': '10.20.30.1'}]
         attrs = {
-            'routes': self._router.routes + [{'destination': '10.20.30.0/24',
-                                             'nexthop': '10.20.30.1'}],
+            'routes': routes + self._router.routes
         }
         self.network.update_router.assert_called_once_with(
             self._router, **attrs)
@@ -731,21 +894,31 @@ class TestSetRouter(TestRouter):
             self._router, **attrs)
         self.assertIsNone(result)
 
-    def test_set_route_no_route(self):
+    def test_set_route_overwrite_route(self):
+        _testrouter = network_fakes.FakeRouter.create_one_router(
+            {'routes': [{"destination": "10.0.0.2",
+                         "nexthop": "1.1.1.1"}]})
+        self.network.find_router = mock.Mock(return_value=_testrouter)
         arglist = [
-            self._router.name,
+            _testrouter.name,
             '--route', 'destination=10.20.30.0/24,gateway=10.20.30.1',
             '--no-route',
         ]
         verifylist = [
-            ('router', self._router.name),
+            ('router', _testrouter.name),
             ('routes', [{'destination': '10.20.30.0/24',
                          'gateway': '10.20.30.1'}]),
             ('no_route', True),
         ]
-
-        self.assertRaises(tests_utils.ParserException, self.check_parser,
-                          self.cmd, arglist, verifylist)
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        result = self.cmd.take_action(parsed_args)
+        attrs = {
+            'routes': [{'destination': '10.20.30.0/24',
+                        'nexthop': '10.20.30.1'}]
+        }
+        self.network.update_router.assert_called_once_with(
+            _testrouter, **attrs)
+        self.assertIsNone(result)
 
     def test_set_clear_routes(self):
         arglist = [
@@ -767,21 +940,31 @@ class TestSetRouter(TestRouter):
             self._router, **attrs)
         self.assertIsNone(result)
 
-    def test_set_route_clear_routes(self):
+    def test_overwrite_route_clear_routes(self):
+        _testrouter = network_fakes.FakeRouter.create_one_router(
+            {'routes': [{"destination": "10.0.0.2",
+                         "nexthop": "1.1.1.1"}]})
+        self.network.find_router = mock.Mock(return_value=_testrouter)
         arglist = [
-            self._router.name,
+            _testrouter.name,
             '--route', 'destination=10.20.30.0/24,gateway=10.20.30.1',
             '--clear-routes',
         ]
         verifylist = [
-            ('router', self._router.name),
+            ('router', _testrouter.name),
             ('routes', [{'destination': '10.20.30.0/24',
                          'gateway': '10.20.30.1'}]),
             ('clear_routes', True),
         ]
-
-        self.assertRaises(tests_utils.ParserException, self.check_parser,
-                          self.cmd, arglist, verifylist)
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        result = self.cmd.take_action(parsed_args)
+        attrs = {
+            'routes': [{'destination': '10.20.30.0/24',
+                        'nexthop': '10.20.30.1'}]
+        }
+        self.network.update_router.assert_called_once_with(
+            _testrouter, **attrs)
+        self.assertIsNone(result)
 
     def test_set_nothing(self):
         arglist = [
@@ -794,16 +977,156 @@ class TestSetRouter(TestRouter):
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
         result = self.cmd.take_action(parsed_args)
 
-        attrs = {}
-        self.network.update_router.assert_called_once_with(
-            self._router, **attrs)
+        self.assertFalse(self.network.update_router.called)
+        self.assertFalse(self.network.set_tags.called)
         self.assertIsNone(result)
 
+    def test_wrong_gateway_params(self):
+        arglist = [
+            "--fixed-ip", "subnet='abc'",
+            self._router.id,
+        ]
+        verifylist = [
+            ('fixed_ip', [{'subnet': "'abc'"}]),
+            ('router', self._router.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        self.assertRaises(exceptions.CommandError,
+                          self.cmd.take_action, parsed_args)
+
+    def test_set_gateway_network_only(self):
+        arglist = [
+            "--external-gateway", self._network.id,
+            self._router.id,
+        ]
+        verifylist = [
+            ('external_gateway', self._network.id),
+            ('router', self._router.id),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.network.update_router.assert_called_with(
+            self._router, **{'external_gateway_info': {
+                'network_id': self._network.id}})
+        self.assertIsNone(result)
+
+    def test_set_gateway_options_subnet_only(self):
+        arglist = [
+            "--external-gateway", self._network.id,
+            "--fixed-ip", "subnet='abc'",
+            self._router.id,
+            '--enable-snat',
+        ]
+        verifylist = [
+            ('router', self._router.id),
+            ('external_gateway', self._network.id),
+            ('fixed_ip', [{'subnet': "'abc'"}]),
+            ('enable_snat', True),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.network.update_router.assert_called_with(
+            self._router, **{'external_gateway_info': {
+                'network_id': self._network.id,
+                'external_fixed_ips': [{
+                    'subnet_id': self._subnet.id, }],
+                'enable_snat': True, }})
+        self.assertIsNone(result)
+
+    def test_set_gateway_option_ipaddress_only(self):
+        arglist = [
+            "--external-gateway", self._network.id,
+            "--fixed-ip", "ip-address=10.0.1.1",
+            self._router.id,
+            '--enable-snat',
+        ]
+        verifylist = [
+            ('router', self._router.id),
+            ('external_gateway', self._network.id),
+            ('fixed_ip', [{'ip-address': "10.0.1.1"}]),
+            ('enable_snat', True),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.network.update_router.assert_called_with(
+            self._router, **{'external_gateway_info': {
+                'network_id': self._network.id,
+                'external_fixed_ips': [{
+                    'ip_address': "10.0.1.1", }],
+                'enable_snat': True, }})
+        self.assertIsNone(result)
+
+    def test_set_gateway_options_subnet_ipaddress(self):
+        arglist = [
+            "--external-gateway", self._network.id,
+            "--fixed-ip", "subnet='abc',ip-address=10.0.1.1",
+            self._router.id,
+            '--enable-snat',
+        ]
+        verifylist = [
+            ('router', self._router.id),
+            ('external_gateway', self._network.id),
+            ('fixed_ip', [{'subnet': "'abc'",
+                           'ip-address': "10.0.1.1"}]),
+            ('enable_snat', True),
+        ]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+
+        result = self.cmd.take_action(parsed_args)
+        self.network.update_router.assert_called_with(
+            self._router, **{'external_gateway_info': {
+                'network_id': self._network.id,
+                'external_fixed_ips': [{
+                    'subnet_id': self._subnet.id,
+                    'ip_address': "10.0.1.1", }],
+                'enable_snat': True, }})
+        self.assertIsNone(result)
+
+    def _test_set_tags(self, with_tags=True):
+        if with_tags:
+            arglist = ['--tag', 'red', '--tag', 'blue']
+            verifylist = [('tags', ['red', 'blue'])]
+            expected_args = ['red', 'blue', 'green']
+        else:
+            arglist = ['--no-tag']
+            verifylist = [('no_tag', True)]
+            expected_args = []
+        arglist.append(self._router.name)
+        verifylist.append(
+            ('router', self._router.name))
+
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        result = self.cmd.take_action(parsed_args)
+
+        self.assertFalse(self.network.update_router.called)
+        self.network.set_tags.assert_called_once_with(
+            self._router,
+            tests_utils.CompareBySet(expected_args))
+        self.assertIsNone(result)
+
+    def test_set_with_tags(self):
+        self._test_set_tags(with_tags=True)
+
+    def test_set_with_no_tag(self):
+        self._test_set_tags(with_tags=False)
+
 
 class TestShowRouter(TestRouter):
 
     # The router to set.
     _router = network_fakes.FakeRouter.create_one_router()
+    _port = network_fakes.FakePort.create_one_port({
+        'device_owner': 'network:router_interface',
+        'device_id': _router.id
+    })
+    setattr(_router,
+            'interfaces_info',
+            [{'port_id': _port.id,
+              'ip_address': _port.fixed_ips[0]['ip_address'],
+              'subnet_id': _port.fixed_ips[0]['subnet_id']}])
 
     columns = (
         'admin_state_up',
@@ -814,10 +1137,12 @@ class TestShowRouter(TestRouter):
         'external_gateway_info',
         'ha',
         'id',
+        'interfaces_info',
         'name',
         'project_id',
         'routes',
         'status',
+        'tags',
     )
     data = (
         router._format_admin_state(_router.admin_state_up),
@@ -825,19 +1150,22 @@ class TestShowRouter(TestRouter):
         osc_utils.format_list(_router.availability_zones),
         _router.description,
         _router.distributed,
-        router._format_external_gateway_info(_router.external_gateway_info),
+        router._format_router_info(_router.external_gateway_info),
         _router.ha,
         _router.id,
+        router._format_router_info(_router.interfaces_info),
         _router.name,
         _router.tenant_id,
         router._format_routes(_router.routes),
         _router.status,
+        osc_utils.format_list(_router.tags),
     )
 
     def setUp(self):
         super(TestShowRouter, self).setUp()
 
         self.network.find_router = mock.Mock(return_value=self._router)
+        self.network.ports = mock.Mock(return_value=[self._port])
 
         # Get the command object to test
         self.cmd = router.ShowRouter(self.app, self.namespace)
@@ -862,6 +1190,9 @@ class TestShowRouter(TestRouter):
 
         self.network.find_router.assert_called_once_with(
             self._router.name, ignore_missing=False)
+        self.network.ports.assert_called_with(**{
+            'device_id': self._router.id
+        })
         self.assertEqual(self.columns, columns)
         self.assertEqual(self.data, data)
 
@@ -874,10 +1205,12 @@ class TestUnsetRouter(TestRouter):
             {'routes': [{"destination": "192.168.101.1/24",
                          "nexthop": "172.24.4.3"},
                         {"destination": "192.168.101.2/24",
-                         "nexthop": "172.24.4.3"}], })
+                         "nexthop": "172.24.4.3"}],
+             'tags': ['green', 'red'], })
         self.fake_subnet = network_fakes.FakeSubnet.create_one_subnet()
         self.network.find_router = mock.Mock(return_value=self._testrouter)
         self.network.update_router = mock.Mock(return_value=None)
+        self.network.set_tags = mock.Mock(return_value=None)
         # Get the command object to test
         self.cmd = router.UnsetRouter(self.app, self.namespace)
 
@@ -915,3 +1248,44 @@ class TestUnsetRouter(TestRouter):
         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
         self.assertRaises(exceptions.CommandError,
                           self.cmd.take_action, parsed_args)
+
+    def test_unset_router_external_gateway(self):
+        arglist = [
+            '--external-gateway',
+            self._testrouter.name,
+        ]
+        verifylist = [('external_gateway', True)]
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        result = self.cmd.take_action(parsed_args)
+        attrs = {'external_gateway_info': {}}
+        self.network.update_router.assert_called_once_with(
+            self._testrouter, **attrs)
+        self.assertIsNone(result)
+
+    def _test_unset_tags(self, with_tags=True):
+        if with_tags:
+            arglist = ['--tag', 'red', '--tag', 'blue']
+            verifylist = [('tags', ['red', 'blue'])]
+            expected_args = ['green']
+        else:
+            arglist = ['--all-tag']
+            verifylist = [('all_tag', True)]
+            expected_args = []
+        arglist.append(self._testrouter.name)
+        verifylist.append(
+            ('router', self._testrouter.name))
+
+        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
+        result = self.cmd.take_action(parsed_args)
+
+        self.assertFalse(self.network.update_router.called)
+        self.network.set_tags.assert_called_once_with(
+            self._testrouter,
+            tests_utils.CompareBySet(expected_args))
+        self.assertIsNone(result)
+
+    def test_unset_with_tags(self):
+        self._test_unset_tags(with_tags=True)
+
+    def test_unset_with_all_tag(self):
+        self._test_unset_tags(with_tags=False)
